/*
 * inetcis.centrald.c
 */

/*
 * System includes
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/wait.h>    /* wait() */
#include <netinet/in.h>  /* Internet socket structures */
#include <fcntl.h>       /* open() */
#include <unistd.h>
#include <netdb.h>
#include <pwd.h>         /* struct passwd, getpwnam() */
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
 
/*
 * Inetcis includes
 */
#include <log.h>
#include <poserr.h>
#include <x25io.h>
#include <misc.h>
#include <command.h>
#include <inetcis-types.h>
#include <config.h>

/*
 * Constants
 */
#define DESIRED_SYSLOG_FD        3
#define DESIRED_SOCK_FD          4
#define DESIRED_X25_FD           5
#define DESIRED_CMD_FD           6
#define MAXSTRLEN                255
#define MAXBUFSIZ                2048
#define COMPUSERVE_PROGRAM_ID   "GCS"

/*
 * Global variables
 */
char Program_Name[MAXSTRLEN+1];

/*
 * Function prototypes
 */
void Exit_Handler ();
void Signal_Handler (int signal_number);
void Setup (char **argv);
void Become_Daemon ();
void Receive_Connections ();
int Setup_Master_Receive_Socket (int desired_number, int *sock_fd);
int Prepare_Tpad_For_Select_Loop (int *x25_fd);
int Setup_New_X25_Device (int desired_number, int *x25_fd,
			  char *x25_device_name);
int Ready_X25_Device (int *x25_fd, char *x25_device_name);
void Deal_With_X25_Problem (int x25_fd);
void Receive_Socket_Connection (int sock_fd, int *peer_sock_fd,
				struct sockaddr_in *peer_address);
void Start_Server (int sock_fd, int peer_sock_fd, int x25_fd,
		   char *x25_device_name);
void Duplicate_On_012 (int fd);
void Get_Socket_Server_File (char *file);
void Move_To_Home_Directory ();
int Process_Incoming_User (int sock_fd, int x25_fd, char *x25_device_name);
int Process_Incoming_Command (int cmd_fd, int sock_fd,
			      int *allowing_connections);
int Process_X25_Problem (int *x25_fd, char *x25_device_name);

int main (int argc, char **argv)
{
  Become_Daemon ();
  Setup (argv);
  Receive_Connections ();
  return EXIT_SUCCESS;     /* not sure if this is the correct thing to do */
}

void
Exit_Handler ()
{
  char host_name[MAXSTRLEN+1];
  int status;

  status = gethostname (host_name, MAXSTRLEN);
  if (status == GETHOSTNAME_ERROR)
    Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
			       "host name", "gethostname", "N/A",
			       "could not get host name");
  else {
    host_name[MAXSTRLEN] = (char) 0;
    if (strncmp(host_name,"mh",2) == 0)
      Write_Message_To_Log (LOG_LEVEL_DEBUG0, __FILE__, __LINE__,
			    "closing log and exiting");
#if 0
    else
      Write_Message_To_Log (LOG_LEVEL_EMERG, __FILE__, __LINE__,
			    "exiting - THIS SHOULD NOT HAPPEN");
#endif
  }

  Close_Log ();
}

void
Signal_Handler (int signal_number)
{
  char signal_number_ascii[10];
  int child_status;
  pid_t child_pid;
  
  switch (signal_number) {
  case SIGCHLD:
    while ((child_pid = wait(&child_status)) == WAIT_ERROR) {
      switch (errno) {
      case ECHILD:
	Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
			      "child", "wait", "N/A", "no child to wait on");
	break;
      case EFAULT:
	Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
				   "child", "wait", "N/A", "wait argument "
				   "invalid");
	exit (EXIT_FAILURE);
      case EINTR:
	continue;
      default:
	break;
      }
    }
    Write_Va_Message_To_Log (LOG_LEVEL_DEBUG0, __FILE__, __LINE__,
			     "child %d returned status %d",
			     child_pid, child_status);
    break;

  case SIGSEGV:
    Write_Error_To_Log (LOG_LEVEL_CRIT, __FILE__, __LINE__, "SIGSEGV",
			"unknown - check core dump", "INE",
			"Gateway software internal error");
    abort ();
    break;
    
  case SIGPIPE:
    Write_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, "SIGPIPE",
			"most likely a disconnect caused by SO_KEEPALIVE",
			"NRT", "No response timeout");
    exit (EXIT_FAILURE);
    break;
    
  default:
    sprintf (signal_number_ascii, "%d", signal_number);
    Write_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__,
			signal_number_ascii,
			"unknown", "UEX", "Unexpected signal caught");
    exit (EXIT_FAILURE);
    break;
  }
}



void
Setup (char **argv)
{
  int status;

  strcpy (Program_Name, argv[0]);

#ifdef LOCAL_LOG_FILE_NAME
  Start_Local_File_Log (LOCAL_LOG_FILE_NAME, "inetcis.centrald");
#endif

#ifndef NO_SYSLOG_LOGGING
  Start_Syslog_Log (Program_Name);
#endif

  status = atexit (Exit_Handler);
  if (status == ATEXIT_ERROR)
    Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
			       "Exit_Handler", "atexit", "N/A",
			       "exit handler setup failed");

  (void) signal (SIGCHLD, Signal_Handler);
#if SIGSEGV_HANDLER
  (void) signal (SIGSEGV, Signal_Handler);
#endif
}

void
Become_Daemon ()
{
  pid_t pid;

  pid = fork ();
  if (pid < 0) {
    Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
			       "N/A", "fork", "N/A", "initial fork failed");
    exit (EXIT_FAILURE);
  } else if (pid != 0) {
    exit (EXIT_SUCCESS);  /* cause parent to exit */
  }

  /*
   * child continues
   */
  setsid ();                  /* become session leader */
  Move_To_Home_Directory ();  /* move to a sane place to generate core files */
  increase_process_limit ();  /* from misc.h */
  umask (0);                  /* duh? */
}

void
Receive_Connections ()
{
  char x25_device_name[MAXSTRLEN+1];
  int sock_fd = -1, x25_fd = -1, cmd_fd;
  int max_fd;
  fd_set reader_fd_set;
  int number_ready;
  int allowing_connections = 1;
  int master_socket_open = 0;
  int x25_device_open = 0;
  int command_socket_open = 0;

  command_socket_open = Create_Server_Command_Socket (DESIRED_CMD_FD, &cmd_fd);

  /***************************************************************************
   *                                                                         *
   *                           Main Select Loop                              *
   *                                                                         *
   ***************************************************************************/
  while (1) {
    FD_ZERO(&reader_fd_set);

    if (! x25_device_open)
      x25_device_open = Setup_New_X25_Device (DESIRED_X25_FD, &x25_fd,
					      x25_device_name);
    if (x25_device_open)
      FD_SET(x25_fd, &reader_fd_set);

    if (allowing_connections && x25_device_open && ! master_socket_open) {
      master_socket_open = Setup_Master_Receive_Socket (DESIRED_SOCK_FD,
							&sock_fd);
    } else if ((! allowing_connections || ! x25_device_open)
	       && master_socket_open) {
      Close_Descriptor (sock_fd);
      master_socket_open = FALSE;
    }

    if (master_socket_open)
      FD_SET(sock_fd, &reader_fd_set);

    if (command_socket_open)
      FD_SET(cmd_fd, &reader_fd_set);

    max_fd = max3 (sock_fd, cmd_fd, x25_fd);
    number_ready = select (max_fd+1, &reader_fd_set,
			   (fd_set *) NULL,
			   (fd_set *) NULL,
			   (struct timeval *) NULL);
    if (number_ready == SELECT_ERROR) {
      if (errno == EINTR) {  /* an exiting child will trigger this */
	Write_System_Error_To_Log (LOG_LEVEL_DEBUG0, __FILE__, __LINE__,
				   errno, "interrupt", "select", "N/A",
				   "select interrupt");
	continue;
      } else {
	Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
				   "unknown", "select", "N/A",
				   "select error");
	exit (EXIT_FAILURE);
      }
    }

    if (FD_ISSET(sock_fd, &reader_fd_set))
      if (Process_Incoming_User (sock_fd, x25_fd, x25_device_name))
	x25_device_open = FALSE;
    
    if (FD_ISSET(cmd_fd, &reader_fd_set))
      Process_Incoming_Command (cmd_fd, x25_fd, &allowing_connections);

    if (FD_ISSET(x25_fd, &reader_fd_set))
      if (Process_X25_Problem (&x25_fd, x25_device_name) == FAILURE)
	x25_device_open = FALSE;
	
  }
  /***************************************************************************
   *                                                                         *
   *                          End Main Select Loop                           *
   *                                                                         *
   ***************************************************************************/
}

int
Setup_Master_Receive_Socket (int desired_number, int *sock_fd)
{
  char port_string[MAXSTRLEN+1];
  struct sockaddr_in sock_address;
  int status;

  bzero (&sock_address, sizeof(struct sockaddr_in));
  *sock_fd = socket (AF_INET, SOCK_STREAM, 0);
  if (*sock_fd == SOCKET_ERROR) {
    Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
			       "socket", "socket", "N/A",
			       "could not open a socket");
    exit (EXIT_FAILURE);
  }

  Position_File_Descriptor (sock_fd, desired_number);

  sock_address.sin_family      = AF_INET;
  sock_address.sin_port        = Convert_Service_To_Port ("inetcis");
  sock_address.sin_addr.s_addr = INADDR_ANY;

  status = bind (*sock_fd, (struct sockaddr *) &sock_address,
		 (int) sizeof(sock_address));
  if (status == BIND_ERROR) {
    /* port number needs to be converted back to host byte order */
    sprintf (port_string, "%d", sock_address.sin_port);
    Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__,
			       errno, port_string , "bind", "N/A",
			       "could not bind to port");
    exit (EXIT_FAILURE);
  }

  status = listen (*sock_fd, 20);
  if (status == LISTEN_ERROR) {
    Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
			       "socket", "listen", "N/A",
			       "problem with master socket");
    exit (EXIT_FAILURE);
  }

  return SUCCESS;
}

int
Prepare_Tpad_For_Select_Loop (int *x25_fd)
{
  if (Position_File_Descriptor (x25_fd, DESIRED_X25_FD) == FAILURE)
    return FAILURE;

  /*
   * The main select statement is going to wait for the X.25 device to
   * time out, but it is known that the X.25 device automatically
   * produces a prompt when it is opened or after Test_X25_Device.
   * So, the following gets the prompt out of the way so the main
   * select loop doesn't have to deal with it.
   */
  if (Read_PAD_Prompt (*x25_fd) == FAILURE)
    return FAILURE;

  return SUCCESS;
}

int
Setup_New_X25_Device (int desired_number, int *x25_fd, char *x25_device_name)
{
  if (Rotor_Open_X25_Device (x25_fd, x25_device_name) == FAILURE) {
    Write_Error_To_Log (LOG_LEVEL_INFO, __FILE__, __LINE__, "X.25 device",
			"Rotor_Open_X25_Device", "N/A",
			"no X.25 devices are available");
    strcpy (x25_device_name, "");
    return FAILURE;
  }

  Write_Va_Message_To_Log (LOG_LEVEL_DEBUG0, __FILE__, __LINE__,
			   "Using X.25 device %s", x25_device_name);

  return Prepare_Tpad_For_Select_Loop (x25_fd);
}

int
Ready_X25_Device (int *x25_fd, char *x25_device_name)
{
  /*
   * Put PAD back into correct state after main select loop is done.
   */
  if (Ask_PAD_For_Prompt (*x25_fd) == FAILURE) {
    Write_Message_To_Log (LOG_LEVEL_DEBUG2, __FILE__, __LINE__,
			  "Ask_PAD_For_Prompt failed in Ready_X25_Device");
    if (Reopen_X25_Device (x25_fd, x25_device_name) == FAILURE) {
      Write_Message_To_Log (LOG_LEVEL_DEBUG2, __FILE__, __LINE__,
			    "Reopen_X25_Device failed in Ready_X25_Device");
      if (Rotor_Open_X25_Device (x25_fd, x25_device_name) == FAILURE) {
	Write_Message_To_Log (LOG_LEVEL_DEBUG2, __FILE__, __LINE__,
			      "Rotor_Open_X25_Device failed in "
			      "Ready_X25_Device");
	return FAILURE;
      }
    }
  }

  if (Refresh_X25_Device (x25_fd, x25_device_name) == SUCCESS)
    return SUCCESS;
  else
    Write_Message_To_Log (LOG_LEVEL_DEBUG2, __FILE__, __LINE__,
			  "Refresh_X25_Device failed in Ready_X25_Device");

  if (Position_File_Descriptor (x25_fd, DESIRED_X25_FD) == SUCCESS)
    return SUCCESS;
  else {
    Write_Message_To_Log (LOG_LEVEL_DEBUG2, __FILE__, __LINE__,
			  "Position_File_Descriptor failed in "
			  "Ready_X25_Device");
    return FAILURE;
  }
} /* Ready_X25_Device */

void
Start_Server (int sock_fd, int peer_sock_fd, int x25_fd, char *x25_device_name)
{
  int child_pid;
  char x25_fd_string[MAXSTRLEN+1];

  child_pid = fork ();
  if (child_pid == FORK_ERROR) {
    Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
			       "server", "fork", "N/A",
			       "could not fork child");
    exit (EXIT_FAILURE);
  }

  if (child_pid == 0) {
    /*
     * We're in the child.
     */
    char server_path_name[MAXSTRLEN+1];
    int status;

    Close_Descriptor (sock_fd);                 /* close master socket */
    Duplicate_On_012 (peer_sock_fd);            /* stdin, stdout, stderr */
    Get_Socket_Server_File (server_path_name);  /* find the executable */
    if (Ready_X25_Device (&x25_fd, x25_device_name) == FAILURE)
      Log_All_Links_In_Use_Error (__FILE__, __LINE__, "Ready_X25_Device");

    Close_Log ();                               /* close all open logs */
    sprintf (x25_fd_string, "%d", x25_fd);
    /*
     * If this call is successful, it does not return.
     */
    status = execl (server_path_name, "inetcisd", "-c", "-s", "-d", "100",
		    "-n", x25_device_name, "-x", x25_fd_string, (char *) 0);
    if (status == EXECL_ERROR) {
      Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
				 server_path_name, "execl", "N/A",
				 "could not execute server");
      exit (EXIT_FAILURE);
    }
  }
  /*
   * Parent returns.
   */
}
  


void Duplicate_On_012 (int fd)
{
  int i;
  int status;

  for (i = 0; i < 3; i++) {
    if (fd == i)
      continue;
    status = dup2 (fd, i);
    if (status == DUP2_ERROR) {
      char fd_string[MAXSTRLEN+1];
      sprintf (fd_string, "%d", fd);
      Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
				 fd_string, "dup2", "N/A",
				 "could not duplicate file descriptor");
      exit (EXIT_FAILURE);
    }
  }
}



void Get_Socket_Server_File (char *file_name)
{
  int status;
  struct stat inetcisd_stat_struct;
  struct passwd *user_passwd_entry_ptr;
  
  strcpy (file_name, "/usr/local/adm/daemon/inetcisd");
  status = stat (file_name, &inetcisd_stat_struct);
  if (status != STAT_ERROR) {
    Write_Va_Message_To_Log (LOG_LEVEL_DEBUG0, __FILE__, __LINE__,
			     "using server %s", file_name);
    return;
  }

  user_passwd_entry_ptr = getpwuid (geteuid());
  if (user_passwd_entry_ptr != GETPWUID_ERROR) {
    strcpy (file_name, user_passwd_entry_ptr->pw_dir);
    strcat (file_name, "/bin/inetcisd");
    status = stat (file_name, &inetcisd_stat_struct);
    if (status != STAT_ERROR) {
      Write_Va_Message_To_Log (LOG_LEVEL_DEBUG0, __FILE__, __LINE__,
			       "using server %s", file_name);
      return;
    }
  }

  user_passwd_entry_ptr = getpwnam ("inetcis");
  if (user_passwd_entry_ptr != GETPWNAM_ERROR) {
    strcpy (file_name, user_passwd_entry_ptr->pw_dir);
    strcat (file_name, "/bin/inetcisd");
    status = stat (file_name, &inetcisd_stat_struct);
    if (status != STAT_ERROR) {
      Write_Va_Message_To_Log (LOG_LEVEL_DEBUG0, __FILE__, __LINE__,
			       "using server %s", file_name);
      return;
    }
  }

  user_passwd_entry_ptr = getpwnam ("bchatfie");
  if (user_passwd_entry_ptr != GETPWNAM_ERROR) {
    strcpy (file_name, user_passwd_entry_ptr->pw_dir);
    strcat (file_name, "/bin/inetcisd");
    status = stat (file_name, &inetcisd_stat_struct);
    if (status != STAT_ERROR) {
      Write_Va_Message_To_Log (LOG_LEVEL_DEBUG0, __FILE__, __LINE__,
			       "using server %s", file_name);
      return;
    }
  }

  /*
   * If execution gets to this point, then no inbound Internet server was
   * found on this machine.  Issue an error and exit.
   */
  Write_Error_To_Log (LOG_LEVEL_CRIT, __FILE__, __LINE__,
		      "inbound Internet Server", "locate", "SRV",
		      "Gateway server is missing");
  exit (EXIT_FAILURE);
}


void Move_To_Home_Directory ()
{
  uid_t uid;
  struct passwd *passwd_entry_ptr;
  int status;

  uid = geteuid();
  passwd_entry_ptr = getpwuid (uid);
  if (passwd_entry_ptr == GETPWUID_ERROR) {
    char uid_error_string[MAXSTRLEN+1];
    sprintf (uid_error_string, "user id %d", uid); 
    Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
			       uid_error_string, "getpwuid", "N/A",
			       "could not get passwd entry for user");
    return;
  }

  status = chdir (passwd_entry_ptr->pw_dir);
  if (status == CHDIR_ERROR) {
    Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
			       passwd_entry_ptr->pw_dir, "chdir", "N/A",
			       "could not change to directory");
    chdir ("/");
    Write_Message_To_Log (LOG_LEVEL_DEBUG0, __FILE__, __LINE__,
			  "made / the working directory");
    return;
  }
  
  Write_Va_Message_To_Log (LOG_LEVEL_DEBUG0, __FILE__, __LINE__,
			   "made %s the working directory",
			   passwd_entry_ptr->pw_dir);
}

int
Process_Incoming_User (int sock_fd, int x25_fd, char *x25_device_name)
{
  struct sockaddr peer_address;
  int peer_sock_fd;
  int length;
  
  length = sizeof (peer_address);
  peer_sock_fd = accept (sock_fd, (struct sockaddr *) &peer_address,
			 &length);
  if (peer_sock_fd == ACCEPT_ERROR) {
    Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
			       "socket", "accept", "N/A",
				   "socket accept failed");
    return FAILURE;
  }
  Start_User_Fd_Log (sock_fd, COMPUSERVE_PROGRAM_ID, "\r\n");
  Start_Server (sock_fd, peer_sock_fd, x25_fd, x25_device_name);
  Stop_User_Fd_Log ();
  Close_Descriptor (x25_fd);
  Close_Descriptor (peer_sock_fd);
  return SUCCESS;
}


int Process_Incoming_Command (int cmd_fd, int sock_fd,
			      int *allowing_connections)
{
  CommandCodeType cmd_code;

  if (Receive_Command (cmd_fd, &cmd_code) == COMMAND_STATUS_FAILURE)
    return FAILURE;

  switch (cmd_code) {
  case COMMAND_STOP_ACCEPTING_CONNECTIONS:
    *allowing_connections = FALSE;
    break;
  case COMMAND_START_ACCEPTING_CONNECTIONS:
    *allowing_connections = TRUE;
    break;
  case COMMAND_SET_MAXIMUM_CONNECTIONS:
    /* set max connections */
    break;
  default:
    break;
  }
  return SUCCESS;
}

int
Process_X25_Problem (int *x25_fd, char *x25_device_name)
{
  char c;
  int num_read;

  if (Read_From_X25_Device (*x25_fd, &c, 1, &num_read) == FAILURE) {
    if (Reopen_X25_Device (x25_fd, x25_device_name) == SUCCESS) {
      if (Prepare_Tpad_For_Select_Loop (x25_fd) == SUCCESS)
	return SUCCESS;
    }
    return FAILURE;
  }

  Write_Va_Message_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__,
			   "Unexpected character %x read from tpad", c);
  return SUCCESS;
}
