/*
 * command.c
 */

/*
 * System includes
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>

/*
 * Local includes
 */
#include <poserr.h>
#include <log.h>
#include <inetcis-types.h>  /* for CentralServerStateType, StatusType */
#include <command.h>

#define MAXSTRLEN 255
#define MAXBUFSIZ 2048

/*
 * Local prototypes
 */
static CommandStatusType Send_Response (int cmd_fd,
					struct sockaddr *cmd_address_ptr,
					CommandStatusType cmd_status);
static CommandStatusType Receive_Response (int cmd_fd,
					   struct sockaddr_in *server_address);

int Create_Server_Command_Socket (int desired_number, int *cmd_fd)
{
  struct sockaddr_in sock_address;
  int status;

  *cmd_fd = socket (AF_INET, SOCK_DGRAM, 0);
  if (*cmd_fd == SOCKET_ERROR) {
    Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
			       "command socket", "socket", "N/A",
			       "could not create command socket");
  }

  sock_address.sin_family      = AF_INET;
  sock_address.sin_port        = Convert_Service_To_Port ("inetcis-cmd");
  sock_address.sin_addr.s_addr = INADDR_ANY;

  status = bind (*cmd_fd, (struct sockaddr *) &sock_address,
		 (int) sizeof(sock_address));
  if (status == BIND_ERROR) {
    char port_string[MAXSTRLEN+1];
    /* %hu means that it is an unsigned short integer */
    sprintf (port_string, "%hu", ntohs(sock_address.sin_port));
    Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__,
			       errno, port_string , "bind", "N/A",
			       "could not bind to port");
    return FAILURE;
  }

  return Position_File_Descriptor (cmd_fd, desired_number);
}

int Create_Client_Command_Socket (int *cmd_fd)
{
  struct sockaddr_in sock_address;
  int status;

  *cmd_fd = socket (AF_INET, SOCK_DGRAM, 0);
  if (*cmd_fd == SOCKET_ERROR) {
    Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
			       "command socket", "socket", "N/A",
			       "could not create command socket");
    return FAILURE;
  }

  return SUCCESS;
}  


int
Send_Simple_Command (int cmd_fd, struct sockaddr_in *server_address,
		     CommandCodeType cmd_code)
{
  int status;

  status = sendto (cmd_fd, &cmd_code, 1, 0, (struct sockaddr *) server_address,
		   sizeof(struct sockaddr_in));
  if (status == SENDTO_ERROR) {
    Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__,
			       errno, "command socket", "sendto", "N/A",
			       "failed to send command");
    return FAILURE;
  }

  return Receive_Response (cmd_fd, server_address);
}  

CommandStatusType Receive_Command (int fd, CommandCodeType *cmd_code)
{
  char buffer[MAXBUFSIZ];
  int number_of_bytes_read;
  struct sockaddr client_address;
  int client_address_length;

  /*
   * Read in the command.
   */
  client_address_length = (int) sizeof (client_address);
  number_of_bytes_read = recvfrom (fd, buffer, MAXBUFSIZ, 0, &client_address,
				   &client_address_length);
  if (number_of_bytes_read == RECVFROM_ERROR) {
    Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
			       "command", "recvfrom", "N/A", "could not read"
			       " command");
    Send_Response (fd, &client_address, COMMAND_STATUS_FAILURE);
    return COMMAND_STATUS_FAILURE;
  }
  Send_Response (fd, &client_address, COMMAND_STATUS_SUCCESS);
  *cmd_code = (CommandCodeType) buffer[0];
  return COMMAND_STATUS_SUCCESS;
}

CommandStatusType Send_Response (int cmd_fd, struct sockaddr *cmd_address_ptr,
				 CommandStatusType cmd_status)
{
  int status;

  status = sendto (cmd_fd, &cmd_status, sizeof(cmd_status), NULL,
		   cmd_address_ptr, sizeof(struct sockaddr));
  if (status == SENDTO_ERROR) {
    Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
			       "command socket", "sendto", "N/A", "could not "
			       "send response through command socket");
    return COMMAND_STATUS_FAILURE;
  }
  return COMMAND_STATUS_SUCCESS;
}

CommandStatusType Receive_Response (int cmd_fd,
				    struct sockaddr_in *server_address)
{
  int status;
  CommandStatusType cmd_status;
  int address_length;

  address_length = sizeof(struct sockaddr_in);
  status = recvfrom (cmd_fd, &cmd_status, sizeof(cmd_status), 0,
		     (struct sockaddr *) server_address,
		     &address_length);

  if (status == RECVFROM_ERROR) {
    Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
			       "command socket", "recvfrom", "N/A",
			       "could not recieve response");
    return COMMAND_STATUS_FAILURE;
  }
  return cmd_status;
}
