/*
 * Copyright (c) 1995 by CompuServe Incorporated, Columbus, Ohio
 * 
 * Facility:  libinetcis.a
 * 
 * Abstract:
 * 
 * This module contains the functions for performing input/output on an X.25
 * device (tpad).  It handles errors and can cause a process to exit. So, the
 * program that uses this module should have an exit handler defined with
 * atexit().  It was designed for inetcis processes.
 * 
 * Environment:  BSDI's BSD/OS 1.1 (POSIX), Requires xpad/tpad interfaces to be
 * available
 * 
 * Author:  William L. Chatfield (BCHATFIELD), 12-Jul-95
 * 
 * Revision History:
 * 
 * 1.0  William L. Chatfield (BCHATFIELD), 12-Jul-95 - Initial revision
 */

/*
 * System External Interfaces
 */
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>           /* for stat() */
#include <sys/ioctl.h>          /* for TIOCEXCL */
#include <unistd.h>		/* for read, write */
#include <errno.h>		/* for errno */
#include <fcntl.h>
#include <syslog.h>		/* for syslog, LOG_ERR, LOG_DEBUG */

/*
 * ANSI C External Interfaces
 */
#include <stdlib.h>		/* for EXIT_SUCCESS, EXIT_FAILURE */
#include <string.h>

/*
 * CompuServe External Interfaces
 */
#include <comdef.h>		/* CompuServe definitions */

/*
 * Project Specific External Interfaces
 */
#include <inetcis-types.h>
#include <x25io.h>
#include <poserr.h>
#include <log.h>

#define MAX_TPAD_DEVICES 128

/*
 * Internal function prototypes
 */
int Open_This_X25_Device (char *x25_device_name, int *x25_fd);

/*
 * Module-wide definitions
 */
static const char     Tpad_Name_Table[MAX_TPAD_DEVICES][3] = {
  /* p */
  "p0", "p1", "p2", "p3", "p4", "p5", "p6", "p7",
  "p8", "p9", "pa", "pb", "pc", "pd", "pe", "pf",
  /* q */
  "q0", "q1", "q2", "q3", "q4", "q5", "q6", "q7",
  "q8", "q9", "qa", "qb", "qc", "qd", "qe", "qf",
  /* r */
  "r0", "r1", "r2", "r3", "r4", "r5", "r6", "r7",
  "r8", "r9", "ra", "rb", "rc", "rd", "re", "rf",
  /* s */
  "s0", "s1", "s2", "s3", "s4", "s5", "s6", "s7",
  "s8", "s9", "sa", "sb", "sc", "sd", "se", "sf",
  /* t */
  "t0", "t1", "t2", "t3", "t4", "t5", "t6", "t7",
  "t8", "t9", "ta", "tb", "tc", "td", "te", "tf",
  /* u */
  "u0", "u1", "u2", "u3", "u4", "u5", "u6", "u7",
  "u8", "u9", "ua", "ub", "uc", "ud", "ue", "uf",
  /* v */
  "v0", "v1", "v2", "v3", "v4", "v5", "v6", "v7",
  "v8", "v9", "va", "vb", "vc", "vd", "ve", "vf",
  /* w */
  "w0", "w1", "w2", "w3", "w4", "w5", "w6", "w7",
  "w8", "w9", "wa", "wb", "wc", "wd", "we", "wf"
  /* add lines for xyzPQRST for 256 devices */
};
static int      produce_flow_stats = 0;
static double   old_number_of_buffers_read = 0.0;
static double   number_of_buffers_read = 0.0;
static double   average_read_buffer_size = 0.0;
static double   number_of_read_eagain_errors = 0.0;
static double   old_number_of_buffers_written = 0.0;
static double   number_of_buffers_written = 0.0;
static double   average_write_buffer_size = 0.0;
static double   number_of_write_eagain_errors = 0.0;

int
Open_This_Tpad_Device (char *tpad_device_name, int *tpad_fd)
{
  *tpad_fd = open (tpad_device_name, O_RDWR | O_NONBLOCK);

  if (*tpad_fd == OPEN_ERROR) {
    /* An EAGAIN error indicates that there is no xpad daemon running
       for the corresponding ppad of this tpad device. */
    if (errno == EAGAIN) {
      Log_Hidden_System_Error (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
			       tpad_device_name, "open",
			       "No xpad running for ppad/tpad device pair");
    } else if (errno != EBUSY) {  /* EBUSY means somebody's already got it. */
      Log_Hidden_System_Error (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
			       tpad_device_name, "open",
			       "Could not open device");
    }
    return FAILURE;
  }

  /* The BSD/386 kernel has been changed so that it defaults to
     exclusive access mode, but if we switch to a new kernel and
     forget to redo the change I want this still to work.  */
  if (ioctl (*tpad_fd, TIOCEXCL, 0) == IOCTL_ERROR) {
    Log_Hidden_System_Error (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
			     tpad_device_name, "ioctl TIOCEXCL",
			     "Could not set exclusive mode for device");
    return FAILURE;
  }				/* if */
  
  return SUCCESS;
}

int 
Open_X25_Device (char *tpad_name)
/*
 * Functional Description:
 * 
 * This function finds an available X.25 device (tpad) on the machine and opens
 * it.
 * 
 * Formal Parameters:
 * 
 * tpad_name (Write only) - This is the variable in which the name of the device
 * that was opened is returned.
 * 
 * Implicit Parameters:  None
 * 
 * Return Value:
 * 
 * If a device could be found, its file descriptor is returned. If no available
 * device was found GET_TPAD_ERROR_NO_AVAILABLE_DEVICE is returned.
 * 
 * Side Effects:  The module-wide variable x25_device_is_alive is set.
 * 
 */
{
  int             fd;
  char           *ptr1, *ptr2;

  strcpy (tpad_name, "/dev/tpadXY");

  /* append xyzPQRST for 256 devices */
  for (ptr1 = "pqrstuvw"; *ptr1 != 0; ptr1++)
  {
    tpad_name[9] = *ptr1;
    for (ptr2 = "0123456789abcdef"; *ptr2 != 0; ptr2++) {
      tpad_name[10] = *ptr2;

      if (Open_This_Tpad_Device (tpad_name, &fd) == FAILURE)
	continue;		/* try next one */

      return (fd);		/* return the file descriptor */
    }				/* inner for */
  }				/* outer for */

  return (GET_TPAD_ERROR_NO_AVAILABLE_DEVICES);	/* out of tpads */
}				/* Open_X25_Device() */

int
Linear_Open_X25_Device (int *tpad_fd, char *tpad_name)
{
  char           *ptr1, *ptr2;

  strcpy (tpad_name, "/dev/tpadXY");

  /*
   * This works for 128 devices.  For 256 add "xyzPQRST" to the ptr1 string.
   */
  for (ptr1 = "pqrstuvw"; *ptr1 != 0; ptr1++) {
    tpad_name[9] = *ptr1;
    for (ptr2 = "0123456789abcdef"; *ptr2 != 0; ptr2++) {
      tpad_name[10] = *ptr2;
      if (Open_This_Tpad_Device (tpad_name, tpad_fd) == FAILURE)
	continue;
      return SUCCESS;
    }				/* inner for */
  }				/* outer for */
  return FAILURE;               /* traversed all devices and didn't find one */
}

int
Reopen_X25_Device (int *x25_fd, char *x25_device_name)
{
  close (*x25_fd);
  sleep (2);

  if (Open_This_Tpad_Device (x25_device_name, x25_fd) == SUCCESS) {
    Write_Va_Message_To_Log (LOG_LEVEL_DEBUG0, __FILE__, __LINE__,
			     "Reopened X.25 device %s on descriptor %d",
			     x25_device_name, *x25_fd);
    return SUCCESS;
  } else {
    Write_Va_Message_To_Log (LOG_LEVEL_DEBUG0, __FILE__, __LINE__,
			     "Unable to reopen device %s", x25_device_name);
    return FAILURE;
  }
}

/* The function Rotor_Open_X25_Device cuts down on the number of open
   system calls that have to be done (over a linear opening scheme) to
   find an available tpad. */

int
Rotor_Open_X25_Device (int *x25_fd, char *x25_device_name)
{
  const char base_name[10] = "/dev/tpad";
  char local_tpad_name[12];
  static int table_index = 0;
  int start_index;
  
  start_index = table_index;  /* Define start point for this invocation. */

  do {
    /* Construct complete tpad device name. */
    strcpy (local_tpad_name, base_name);
    strcat (local_tpad_name, Tpad_Name_Table[table_index]);

    /* Increment table index for next iteration or next invocation. */
    if (++table_index > (MAX_TPAD_DEVICES - 1))
      table_index = 0;
    
    /* Open it and make sure it is going to work. */
    if (Open_This_Tpad_Device (local_tpad_name, x25_fd) == SUCCESS)
      if (Test_X25_Device (*x25_fd) == SUCCESS) {
	strcpy (x25_device_name, local_tpad_name);
	return SUCCESS;
      }
  } while (table_index != start_index);
  
  return FAILURE;  /* There are no working X.25 tpads available. */
}


void 
Start_Producing_X25_Flow_Stats ()
{
  produce_flow_stats = 1;
}

int 
Read_From_X25_Device (int fd, char *buffer, int number_of_bytes_to_read,
		      int *number_of_bytes_read)
/*
 * The following comments may be completely wrong.
 */
/*
 * Functional Description:
 * 
 * This function performs a read() on the file descriptor that is passed to it.
 * It checks for errors and handles them correctly.
 * 
 * Requirements:
 * 
 * 1. There must be a syslog type log file open so that syslog calls will work.
 * 2. fd must be a valid open file descriptor for an X.25 device. 3. fd
 * should be in non-blocking mode.
 * 
 * Formal Parameters:
 * 
 * fd (Read only) - The file descriptor of the X.25 device from which data will
 * be read. buffer (Write only) - A pointer to the area of memory into which
 * the data read will be put.  The size of this area of memory should be
 * equal to number_of_bytes_to_read or greater. number_of_bytes_to_read (Read
 * only) - The number of bytes to read from the X.25 device.  This number
 * should be equal to or less than the size of the memory pointed to by
 * buffer.
 * 
 * Implicit Parameters:  None
 * 
 * Return Values:
 * 
 * The number of bytes read is returned.
 * 
 * Side Effects:  The module-wide variable x25_device_is_alive is set.
 * 
 */
{
  int             n;

  if (number_of_bytes_read == NULL || buffer == NULL)
    return FAILURE;

  n = read (fd, buffer, number_of_bytes_to_read);

  if (n == READ_ERROR) {
    if (errno == EAGAIN) {
      n = 0;
      Log_Hidden_System_Error (LOG_LEVEL_DEBUG0, __FILE__, __LINE__, errno,
			       "X.25 device", "read", "Nothing to read");
      number_of_read_eagain_errors++;
    }
    /*
     * The following case deals with the fact that the X.25 device sends an
     * EIO when the connection closes.
     */
    else if (errno == EIO) {
      Write_System_Error_To_Log (LOG_LEVEL_DEBUG0, __FILE__, __LINE__, errno,
				 "X.25 device", "read", "N/A",
				 "X.25 connection has closed");
      return FAILURE;
    } else {
      Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
				 "X.25 device", "read", "XRD",
				 "Unexpected read error");
      return FAILURE;
    }
  } else if (n == READ_EOF) {
    Write_Error_To_Log (LOG_LEVEL_INFO, __FILE__, __LINE__, "X.25 device",
			"read", "N/A", "EOF read");
    return FAILURE;
  }
  /*
   * Calculate statistics
   */
  if (produce_flow_stats) {
    old_number_of_buffers_read = number_of_buffers_read;
    number_of_buffers_read++;
    average_read_buffer_size = (average_read_buffer_size
				* old_number_of_buffers_read
				+ n)
      / number_of_buffers_read;
  }

  Log_Buffer (__FILE__, __LINE__, buffer, n,
	      "Bytes read from X.25 device");
    
  *number_of_bytes_read = n;
  return SUCCESS;
}				/* Read_from_X25_Device */



int 
Write_To_X25_Device (int fd, char *buffer, int number_of_bytes_to_write,
		     int *number_of_bytes_written)
/*
 * Functional Description:
 * 
 * This function performs a write() on the file descriptor that is passed to it.
 * It checks for errors and handles them correctly.
 * 
 * Requirements:
 * 
 * 1. There must be a syslog type log file open so that syslog calls will work.
 * 2. fd must be a valid open file descriptor for an X.25 device. 3. fd
 * should be in non-blocking mode.
 * 
 * Formal Parameters:
 * 
 * fd (Read only) - The file descriptor of the X.25 device from which data will
 * be read. buffer (Read only) - A pointer to the area of memory from which
 * the data will be taken.  The size of this area of memory should be equal
 * to number_of_bytes_to_read or greater. number_of_bytes_to_write (Read
 * only) - The number of bytes to write to the X.25 device.  This number
 * should be equal to or less than the size of the memory pointed to by
 * buffer.
 * 
 * Implicit Parameters:  None
 * 
 * Return Values:
 *    0 - A fatal error occured
 *    1 - Everything's OK
 * 
 * Side Effects:  
 * 
 */
{
  if (number_of_bytes_written == NULL || buffer == NULL)
    return FAILURE;

  *number_of_bytes_written = write (fd, buffer, number_of_bytes_to_write);

  if (WRITE_ERROR == *number_of_bytes_written) {
    if (errno == EAGAIN) {
      *number_of_bytes_written = 0;
      Log_Hidden_System_Error (LOG_LEVEL_DEBUG0, __FILE__, __LINE__, errno,
			       "X.25 device", "write", "Write would block");
      number_of_write_eagain_errors++;
    } else {
      Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
				 "X.25 device", "write", "WRT",
				 "Problem with write");
      return FAILURE;
    }
  }				/* if (WRITE_ERROR ==
				 * number_of_bytes_written) */
  /*
   * Calculate statistics
   */
  if (produce_flow_stats) {
    old_number_of_buffers_written = number_of_buffers_written;
    number_of_buffers_written++;
    average_write_buffer_size = (average_write_buffer_size
				 * old_number_of_buffers_written
				 + *number_of_bytes_written)
      / number_of_buffers_written;
  }

  Log_Buffer (__FILE__, __LINE__, buffer, *number_of_bytes_written,
	      "Bytes written to X.25 device");

  return SUCCESS;
}				/* Write_to_X25_Device */


int 
Write_All_To_X25_Device (int file_descriptor, char *buffer_ptr,
			 int number_of_bytes_to_write)
/*
 * Functional Description:
 * 
 * This function writes all the bytes specified and does not return until it has
 * written all of them or an error occurs.
 * 
 * Formal Parameters:
 * 
 * file_descriptor (Read only) - descriptor to write to buffer_ptr (Read only) -
 * pointer to the buffer from which the bytes written will be taken
 * number_of_bytes_to_write (Read only) - number of bytes to write
 * 
 * Implicit Parameters:  x25_device_is_alive
 * 
 * Return Value:
 * 
 * None
 * 
 * Side Effects:  None
 * 
 */
{
  int             number_of_bytes_written = 0;
  int             total_bytes_written = 0;
  fd_set          output_set;
  int             select_status;
  struct timeval  timeout;

  timeout.tv_sec = 2;
  timeout.tv_usec = 0;

  while (total_bytes_written < number_of_bytes_to_write) {
    FD_ZERO (&output_set);
    FD_SET (file_descriptor, &output_set);

    select_status = select (file_descriptor + 1, (fd_set *) NULL,
			    &output_set, (fd_set *) NULL, &timeout);

    if (select_status == SELECT_ERROR) {
      syslog (LOG_ERR, "%s %d - err - function select caused system"
	      " error `%m' trying to wait for X.25 device to be ready"
	      " to write", __FILE__, __LINE__);
      if (errno == EINTR)
	continue;
      return 0;
    } else if (select_status == SELECT_TIMEOUT) {
      syslog (LOG_ERR, "%s %d - err - select timeout trying to wait for"
	      " X.25 device to be ready to write", __FILE__, __LINE__);
      return 0;
    } else if (select_status > 1 || !FD_ISSET (file_descriptor, &output_set)) {
      syslog (LOG_ERR, "%s %d - err - unknown select error trying to"
	      " wait for"
	      " X.25 device to be ready to write", __FILE__, __LINE__);
      return 0;
    }
    if (! Write_To_X25_Device (file_descriptor, buffer_ptr,
			       number_of_bytes_to_write - total_bytes_written,
			       &number_of_bytes_written))
      return 0;
    buffer_ptr += number_of_bytes_written;
    total_bytes_written += number_of_bytes_written;
  }
}

int
Write_All_To_X25_Device_EonF (int fd, char *buffer_ptr, int number_to_write)
{
  if (Write_All_To_X25_Device (fd, buffer_ptr, number_to_write) == SUCCESS)
    return SUCCESS;
  exit (EXIT_FAILURE);
}

void 
Log_X25_Device_Stats ()
{
  Write_Va_Message_To_Log (LOG_LEVEL_INFO, __FILE__, __LINE__,
			   "X25 Device Stats: "
			   "Average Read Buffer Size = %f, "
			   "Number of EAGAIN Read Errors = %f, "
			   "Average Write Buffer Size = %f, "
			   "Number of EAGAIN Write Errors = %f",
			   average_read_buffer_size,
			   number_of_read_eagain_errors,
			   average_write_buffer_size,
			   number_of_write_eagain_errors);
}

int
Read_PAD_Prompt (int x25_fd)
{
  char            ast;
  int             system_status;
  fd_set          readers;
  struct timeval  timeout = {5 /* seconds */, 0 /* microseconds */};
  int             n;  /* number of characters read */
  
  do {
    FD_ZERO (&readers);
    FD_SET (x25_fd, &readers);

    system_status = select (x25_fd + 1, &readers, (fd_set *) NULL,
			    (fd_set *) NULL, (struct timeval *) &timeout);
    if (system_status == SELECT_ERROR) {
      Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
				 "X.25 device", "select", "TSL",
				 "Problem waiting for network device");
      return FAILURE;
    } else if (system_status == SELECT_TIMEOUT) {
      Write_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, "X.25 device",
			  "select", "TTO", "Network device timed out");
      return FAILURE;
    }

    if (! Read_From_X25_Device (x25_fd, &ast, 1, &n))
      return FAILURE;
  } while (ast != '*');
}

int
Ask_PAD_For_Prompt (int x25_fd)
{
  return Write_All_To_X25_Device (x25_fd, "\r", 1);
}

/*
 * Function: Test_X25_Device
 *
 * Notes:
 * This function should only be called before a call has been made because
 * it works by trying to read the asterisk prompt.
 */

int 
Test_X25_Device (int x25_fd)
{
  if (Read_PAD_Prompt (x25_fd) == FAILURE)
    return FAILURE;

  if (Ask_PAD_For_Prompt (x25_fd) == FAILURE)
    return FAILURE;

  return SUCCESS;
}

int
Refresh_X25_Device (int *x25_fd, char *x25_device_name)
{
  if (Test_X25_Device (*x25_fd) == SUCCESS)
    return SUCCESS;

  if (Reopen_X25_Device (x25_fd, x25_device_name) == SUCCESS)
    return SUCCESS;

  return Rotor_Open_X25_Device (x25_fd, x25_device_name);
}

void
Log_All_Links_In_Use_Error (char *file, int line, char *operation)
{
  Write_Error_To_Log (LOG_LEVEL_ERR, file, line, "X.25 device", operation,
		      "LNK", "All links to the CompuServe network are in use");
}


/* This function, Read_User_ID_Prompt, has to be used.  If the main
   select loop is entered before the User ID prompt is manually read,
   an error could happen and the user could end up getting an asterisk
   prompt.  That would be bad.

   The matching of prompts is not done quite right here.  If a prompt
   comes across buffers it won't work.  Fortunately, this doesn't
   happen very often. */

void 
Read_User_ID_Prompt (int file_descriptor)
{
  char            stripped_buffer[MAXSTRLEN+1] = {0};  /* define and init */
  char            xpad_message[MAXSTRLEN+1] = {0};  /* define and init */
  char           *newline_ptr;
  int             character_count;
  char            buffer[MAXSTRLEN+1] = {0};  /* define and init */
  int             stripped_buffer_length = 0;
  fd_set          select_readers;
  int             system_status;
  struct timeval  timeout_period = {30L /* seconds */ ,
				      0L /* microseconds */ };

  do {
    FD_ZERO (&select_readers);
    FD_SET (file_descriptor, &select_readers);
    system_status = select (file_descriptor + 1, &select_readers,
			    (fd_set *) NULL,
			    (fd_set *) NULL, &timeout_period);
    if (system_status == SELECT_ERROR) {
      Write_System_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, errno,
				 "tpad", "select", "EUI",
				 "Gateway error waiting for User ID prompt");
      exit (EXIT_FAILURE);
    } else if (system_status == SELECT_TIMEOUT) {
      Write_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, "tpad", "select",
			  "TUI",
			  "Gateway timed out waiting for User ID prompt");
      exit (EXIT_FAILURE);
    } else if (FD_ISSET (file_descriptor, &select_readers)) {
      if (Read_From_X25_Device (file_descriptor, buffer, MAXSTRLEN,
				&character_count) == FAILURE)
	exit (EXIT_FAILURE);
      Strip_Leading_CRs_And_LFs (buffer, stripped_buffer,
				 character_count,
				 &stripped_buffer_length);
    } else {
      Write_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, "tpad", "select",
			  "FUI", "Gateway failed to get User ID prompt");
      exit (EXIT_FAILURE);
    }

    /* Check for an asterisk prompt. */
    if (stripped_buffer[0] == '*') {

      /* Check to see if we already got a message from xpad on a
	 previous loop iteration. */

      if (strlen (xpad_message) != 0)
	Write_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, "xpad",
			    "X.25 network call", "PAD", xpad_message);
      else
	Write_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, "xpad",
			    "X.25 network call", "GXF",
			    "Gateway X.25 network access failure");

      exit (EXIT_FAILURE);
    }

    /* Check for messages from xpad.  ARG! */
    stripped_buffer[stripped_buffer_length] = '\0';  /* safe - MAXSTRLEN+1 */
    if (strncmp (stripped_buffer, "xpad:", 5) == 0) {
      strcpy (xpad_message, stripped_buffer + 5);  /* skip "xpad:" */
      Trim_String (xpad_message);
      /* Check to see if we read two lines at the same time. */
      if ((newline_ptr = strpbrk (xpad_message, "\r\n")) != NULL) {
	*newline_ptr = '\0';  /* Make xpad_message end at newline_ptr. */
	/* Check for asterisk prompt in second line. */
	if (strpbrk (newline_ptr + 1, "*") != NULL) {
	  /* If we already got the asterisk prompt back, print error now. */
	  Write_Error_To_Log (LOG_LEVEL_ERR, __FILE__, __LINE__, "xpad",
			      "X.25 network call", "PAD", xpad_message);
	  exit (EXIT_FAILURE);
	}
      }
    }
    
  } while (stripped_buffer_length <= 0 ||
	   0 != strncmp("User ID: ",stripped_buffer,stripped_buffer_length));
}
